/* MS3DMODEL.CPP - Programmed by Jonathan Hobson */
/* 090404: Maarten Hofman: Allow ASCII 3D model loading */
/* 120404: Maarten Hofman: Add the has_name() function */
/* 060504: Maarten Hofman: Changed light a bit lighter */
/* 090504: Maarten Hofman: And added a third light direction */
/* 180504: Maarten Hofman: Added some checks while loading */
/* 010604: Maarten Hofman: Fixed bug by changing 27 to 26 */
/* 280604: Maarten Hofman: Hunt for the next bug */
/* 230704: Maarten Hofman: Limit initialisation to one time */
/* 280205: Maarten Hofman: Added limited light code */
/* 020305: Maarten Hofman: Changed transformation matrix to full thing */
/* 030305: Maarten Hofman: Set the point light stuff */
/* 060305: Maarten Hofman: Added a style parameter */
/* 090405: Maarten Hofman: Defined style 8 */

#include "ms3dmodel.h"
#include "sps2wrap.h"
#include "dma.h"
#include <memory.h>
#include "light.h"
#include "ps2maths.h"

using namespace KazMaths;

// #define MSVERBOSE

CMs3dModel::CMs3dModel(void)
:m_pMeshes(NULL)
{
}

CMs3dModel::~CMs3dModel(void)
{
	if(m_pMeshes != NULL)
	{
		delete [] m_pMeshes;
		m_pMeshes = NULL;
	}
}

int CMs3dModel::has_name(char *how) {
  return !strcmp(name,how);
}

void CMs3dModel::GetNextLOC(char *pStr1024)
{
	bool foundcode = false;

	while(foundcode != true)
	{
		fgets(pStr1024, 1024, m_fp);

		if((pStr1024[0]!='/')&&(pStr1024[1]!='/')&&(pStr1024[0]!='\r')
		   &&(pStr1024[0]!='\n')) {
                   foundcode=true;
                }
	}
}

bool CMs3dModel::Load(const char *strFilename,int style) {
        ourstyle=style;
        rr=0.0f;
	m_fp = fopen(strFilename, "ra");

	if(m_fp == NULL)
	{
		fprintf(stderr,"Failed to load model %s.\n", strFilename);
                return false;
	}

        strncpy(name,strFilename,32);

	char strLine[1024];

	// Frames
	GetNextLOC(strLine);
	sscanf(strLine, "Frames: %d", &m_nFrames);

#ifdef MSVERBOSE
	printf("Model: %s\nFrames: %d\n", strFilename, m_nFrames);
#endif

	// Current frame
	GetNextLOC(strLine);
	sscanf(strLine, "Frame: %d", &m_nCurrentFrame);

#ifdef MSVERBOSE
	printf("Current Frame: %d\n", m_nCurrentFrame);
#endif

	// No. meshes
	GetNextLOC(strLine);
	sscanf(strLine, "Meshes: %d", &m_nMeshes);

#ifdef MSVERBOSE
	printf("Meshes: %d\n", m_nMeshes);
#endif

	m_pMeshes = new MSMesh[m_nMeshes];
	memset(m_pMeshes, 0, m_nMeshes * sizeof(MSMesh));

	for(int nMesh = 0; nMesh < m_nMeshes; nMesh++)
	{
		GetNextLOC(strLine);
		// mesh: name, flags, material index
		sscanf(strLine,	"\"%[^\"]\" %d %d", m_pMeshes[nMesh].strName,
						&m_pMeshes[nMesh].flags,
						&m_pMeshes[nMesh].matindex);

#ifdef MSVERBOSE
		printf("%s %d %d\n",		m_pMeshes[nMesh].strName,
						m_pMeshes[nMesh].flags,
						m_pMeshes[nMesh].matindex);
#endif

		// No. vertices
		GetNextLOC(strLine);
		sscanf(strLine, "%d", &m_pMeshes[nMesh].nVerts);

#ifdef MSVERBOSE
		printf("Vertices: %d\n", m_pMeshes[nMesh].nVerts);
#endif

		m_pMeshes[nMesh].pVertices = new MSVertex[m_pMeshes[nMesh].nVerts];

		// The vertices
		for(int nVert = 0; nVert < m_pMeshes[nMesh].nVerts; nVert++)
		{
			GetNextLOC(strLine);
			// flags, x, y, z, u, v, bone index
			sscanf(strLine, "%d %f %f %f %f %f %d", 
				&m_pMeshes[nMesh].pVertices[nVert].flags,
				&m_pMeshes[nMesh].pVertices[nVert].x,
				&m_pMeshes[nMesh].pVertices[nVert].y,
				&m_pMeshes[nMesh].pVertices[nVert].z,
				&m_pMeshes[nMesh].pVertices[nVert].u,
				&m_pMeshes[nMesh].pVertices[nVert].v,
				&m_pMeshes[nMesh].pVertices[nVert].bone);

#ifdef MSVERBOSE
			printf("%d %f %f %f %f %f %d\n", 
				m_pMeshes[nMesh].pVertices[nVert].flags,
				m_pMeshes[nMesh].pVertices[nVert].x,
				m_pMeshes[nMesh].pVertices[nVert].y,
				m_pMeshes[nMesh].pVertices[nVert].z,
				m_pMeshes[nMesh].pVertices[nVert].u,
				m_pMeshes[nMesh].pVertices[nVert].v,
				m_pMeshes[nMesh].pVertices[nVert].bone);
#endif
		}

		// No. normals
		GetNextLOC(strLine);
		sscanf(strLine, "%d", &m_pMeshes[nMesh].nNormals);

#ifdef MSVERBOSE
		printf("Normals: %d\n", m_pMeshes[nMesh].nNormals);
#endif

		m_pMeshes[nMesh].pNormals = new KazMaths::Vector3[m_pMeshes[nMesh].nNormals];

		// The normals
		for(int nNormal = 0; nNormal < m_pMeshes[nMesh].nNormals; nNormal++)
		{
			GetNextLOC(strLine);

			// x, y, z
			sscanf(strLine, "%f %f %f",
				&m_pMeshes[nMesh].pNormals[nNormal].x,
				&m_pMeshes[nMesh].pNormals[nNormal].y,
				&m_pMeshes[nMesh].pNormals[nNormal].z);

#ifdef MSVERBOSE
			printf("%f %f %f\n",
				m_pMeshes[nMesh].pNormals[nNormal].x,
				m_pMeshes[nMesh].pNormals[nNormal].y,
				m_pMeshes[nMesh].pNormals[nNormal].z);
#endif
		}

		// No. triangles
		GetNextLOC(strLine);
		sscanf(strLine, "%d", &m_pMeshes[nMesh].nTriangles);

#ifdef MSVERBOSE
		printf("Triangles: %d\n", m_pMeshes[nMesh].nTriangles);
#endif

		m_pMeshes[nMesh].pTriangles = new MSTriangle[m_pMeshes[nMesh].nTriangles];

		// The triangles
		for(int nTriangle = 0; nTriangle < m_pMeshes[nMesh].nTriangles; nTriangle++)
		{
			GetNextLOC(strLine);

			// triangle: flags, vertex index1, vertex index2, vertex index3, 
			// normal index1, normal index 2, normal index 3, smoothing group
			sscanf(strLine, "%d %d %d %d %d %d %d %d",
				&m_pMeshes[nMesh].pTriangles[nTriangle].flags,
				&m_pMeshes[nMesh].pTriangles[nTriangle].vert1,
				&m_pMeshes[nMesh].pTriangles[nTriangle].vert2,
				&m_pMeshes[nMesh].pTriangles[nTriangle].vert3,
				&m_pMeshes[nMesh].pTriangles[nTriangle].norm1,
				&m_pMeshes[nMesh].pTriangles[nTriangle].norm2,
				&m_pMeshes[nMesh].pTriangles[nTriangle].norm3,
				&m_pMeshes[nMesh].pTriangles[nTriangle].smoothinggroup);

#ifdef MSVERBOSE
			printf("%d %d %d %d %d %d %d %d\n",
				m_pMeshes[nMesh].pTriangles[nTriangle].flags,
				m_pMeshes[nMesh].pTriangles[nTriangle].vert1,
				m_pMeshes[nMesh].pTriangles[nTriangle].vert2,
				m_pMeshes[nMesh].pTriangles[nTriangle].vert3,
				m_pMeshes[nMesh].pTriangles[nTriangle].norm1,
				m_pMeshes[nMesh].pTriangles[nTriangle].norm2,
				m_pMeshes[nMesh].pTriangles[nTriangle].norm3,
				m_pMeshes[nMesh].pTriangles[nTriangle].smoothinggroup);
#endif
		}
	}

	fclose(m_fp);

	InitialiseDMA(style);

	return true;
}

void CMs3dModel::InitialiseDMA(int style) {
bool bTransparent;
float uvw;

        uvw=0.0f;
        if(style==2) uvw=0.125f;
        bTransparent=false;
        if(style&1) bTransparent=true;
	m_dmaModel = VIFStaticDMA.GetPointer();

	int iNumTrisPerChunk = 27;
	MSMesh & pMesh = m_pMeshes[0];
	int iNumTris = pMesh.nTriangles;
	int iAbsoluteTri = 0;

	while(iNumTris > 0)
	{
		int iTrisThisChunk;

		if(iNumTris > iNumTrisPerChunk)
		{
			iTrisThisChunk = iNumTrisPerChunk;
			iNumTris -= iTrisThisChunk;
		}
		else
		{
			iTrisThisChunk = iNumTris;
			iNumTris -= iTrisThisChunk;
		}

		VIFStaticDMA.AddUnpack(V4_32, 0, iTrisThisChunk * 9 + 2, 1);

		VIFStaticDMA.Add128(iTrisThisChunk * 9);

		VIFStaticDMA.Add128(
			GS_GIFTAG_BATCH(iTrisThisChunk * 3, // NLOOP
					1,	// EOP
					1,	// PRE
					// GS_PRIM(    PRIM, IIP, TME, FGE,  ABE         , AA1, FST, CTXT, FIX)
					GS_PRIM(GS_TRIANGLE,  1,   1,   0 , bTransparent ,   0,   0,    0,   0),	// PRIM
					GIF_FLG_PACKED,
					GS_BATCH_3(GIF_REG_ST, GIF_REG_RGBAQ, GIF_REG_XYZ2)));

		for(int iTriangle = 0; iTriangle < iTrisThisChunk; iTriangle++)
		{
			MSTriangle & pTriangle = pMesh.pTriangles[iAbsoluteTri++];

			MSVertex & pVert1 = pMesh.pVertices[pTriangle.vert1];
			MSVertex & pVert2 = pMesh.pVertices[pTriangle.vert2];
			MSVertex & pVert3 = pMesh.pVertices[pTriangle.vert3];

			Vector3 & pNorm1 = pMesh.pNormals[pTriangle.norm1];
			Vector3 & pNorm2 = pMesh.pNormals[pTriangle.norm2];
			Vector3 & pNorm3 = pMesh.pNormals[pTriangle.norm3];

			VIFStaticDMA.AddVector(Vector4(pVert1.u,pVert1.v,1,uvw));
			VIFStaticDMA.AddVector(Vector4(pNorm1.x,pNorm1.y,pNorm1.z, 0));
			VIFStaticDMA.AddVector(Vector4(pVert1.x, pVert1.y, pVert1.z, 1));
		
			VIFStaticDMA.AddVector(Vector4(pVert2.u, pVert2.v,1,uvw));
			VIFStaticDMA.AddVector(Vector4(pNorm2.x, pNorm2.y, pNorm2.z, 0));
			VIFStaticDMA.AddVector(Vector4(pVert2.x, pVert2.y, pVert2.z, 1));
			VIFStaticDMA.AddVector(Vector4(pVert3.u,pVert3.v,1,uvw));
			VIFStaticDMA.AddVector(Vector4(pNorm3.x, pNorm3.y, pNorm3.z, 0));
			VIFStaticDMA.AddVector(Vector4(pVert3.x, pVert3.y, pVert3.z, 1));
		}
		
		VIFStaticDMA.Add32(MSCALL(0));
	}

	VIFStaticDMA.Add32(FLUSH);

	VIFStaticDMA.DMARet();
}

void CMs3dModel::Render(const KazMaths::Matrix4x4 & matTrans, const KazMaths::Matrix4x4 & matWorld) {
float frr;

  frr=0.0f;
  if(ourstyle&4) {
    rr+=0.001f;
    if(rr>0.25f) rr=0.0f;
    frr=rr;
  }
  if(ourstyle&8) {
    rr+=0.01f;
    if(rr>6.28f) rr=0.0f;
    frr=(1.0f+Sin(rr))/8.0f;
  }
  VIFDynamicDMA.AddUnpack(V4_32, 8, 10);
  VIFDynamicDMA.AddMatrix(matTrans);
  VIFDynamicDMA.AddMatrix(matWorld);
  if(pointLight) {
    VIFDynamicDMA.AddVector(Vector4(plred,plgreen,plblue,frr));
  }
  else {
    VIFDynamicDMA.AddVector(Vector4(0,0,0,frr));
  }
  get_point_light();
  VIFDynamicDMA.AddVector(Vector4(plx,ply,plz,rand()));
  VIFDynamicDMA.DMACall(m_dmaModel);
}
