#include "ps2vector3.h"

using namespace KazMaths;

Vector3::Vector3(void)
{
}

Vector3::Vector3(const float _x, const float _y, const float _z)
:x(_x),y(_y),z(_z)
{
}

Vector3::Vector3(const Vector3 &rhs)
:x(rhs.x),y(rhs.y),z(rhs.z)
{
}

Vector3::~Vector3(void)
{
}

Vector3 & Vector3::operator +=(const Vector3 &rhs)
{
	x += rhs.x;
	y += rhs.y;
	z += rhs.z;

	return *this;
}

Vector3 & Vector3::operator -=(const Vector3 &rhs)
{
	x -= rhs.x;
	y -= rhs.y;
	z -= rhs.z;

	return *this;
}

float Vector3::Dot(const Vector3 &rhs) const
{
	return (x * rhs.x + y * rhs.y + z * rhs.z);
}

Vector3 Vector3::Cross(const Vector3 &rhs) const
{
	return Vector3(y * rhs.z - z * rhs.y, z * rhs.x - x * rhs.z, x * rhs.y - y * rhs.x); 
}

Vector3 & Vector3::operator *=(const float s)
{
	x *= s;
	y *= s;
	z *= s;

	return *this;
}

Vector3 & Vector3::operator /=(const float s)
{
	x /= s;
	y /= s;
	z /= s;

	return *this;
}

float Vector3::Length() const
{
	return Sqrt(x * x + y * y + z * z);
}

float Vector3::LengthSqr() const
{
	return (x * x + y * y + z * z);
}

Vector3 Vector3::Normalise()
{
	return (*this / this->Length());
}

void Vector3::NormaliseSelf()
{
	*this /= this->Length();
}

Vector3 Vector3::ProjectOnto(const Vector3 & rhs) const
{
	// P = this, Q = rhs
	// Proj = Q(P.Q / |Q|^2)
	float fQlensq = rhs.Length();
	fQlensq *= fQlensq;

	return rhs * (this->Dot(rhs) / fQlensq);
}

Vector3 Vector3::PerpendicularOnto(const Vector3 & rhs) const
{
	return (*this - this->ProjectOnto(rhs));
}

bool Vector3::operator ==(const Vector3 & rhs) const
{
	return ((x == rhs.x) && (y == rhs.y) && (z == rhs.z));
}
