/* WINDOW.CPP - Programmed by Maarten Hofman */
/* 010504: started programming */
/* 020504: Fixed scrolling */
/* 030504: Added the clear option */
/* 080504: Select option */
/* 090504: Added the option to get the selected option */
/* 030804: Added cycle to render option */
/* 100305: Removed debug statements */

#include <string.h>
#include <stdio.h>
#include "window.h" 

window::window() {
  x1=-1;
  y1=-1;
  x2=-2;
  y2=-2;
  /* Why recode everything? Let's use clear() */
  clear();
}

window::~window() {
}

void window::clear() {
  current=0;
  cursor=0;
  scrolling=false;
  scroll=0;
  sstart=-1;
  send=-1;
  scurrent=-1;
  speed=0;
  /* No symbols in list anymore */
  csym=0;
}

void window::add_symbol(int x,int y,int symbol) {
  symbols[csym]=symbol;
  symx[csym]=x;
  symy[csym]=y;
  csym++;
}

void window::add_line(char *which) {
int total;

  strncpy(lines[current],which,MAX_CHARS);
  lines[current][MAX_CHARS-1]=0;
  current++;
  if(current>=MAX_LINES) current=0;
  /* Do not check for scrolling in case we are in select mode */
  if(sstart>=0) return;
  if(!scrolling) {
    total=current-cursor;
    if(total<0) total+=MAX_LINES;
    if(total*16>(y2-y1)) scrolling=true;
  }
}

char *window::get_selected_line() {
  if(sstart<0) return NULL;
  return lines[scurrent];
}

void window::set_size(int nx1,int ny1,int nx2,int ny2) {
  x1=nx1;
  y1=ny1;
  x2=nx2;
  y2=ny2;
}

void window::set_select(int start,int end) {
  sstart=start;
  send=end;
  scurrent=start;
}

void window::move_select(int how) {
  if(sstart<0) return;
  scurrent+=how;
  if(scurrent<sstart) scurrent=send;
  if(scurrent>send) scurrent=sstart;
}

void window::render(CFont *font,int cycle) {
int loc,cy,es,i;
char symbol[2];
bool wa;

  if(x1>x2) return;
  if(cursor==current) return;
  /* Cycle two involves only stationary graphics */
  if(cycle==2) {
    symbol[1]=0;
    for(i=0;i<csym;i++) {
      symbol[0]=symbols[i];
      font->Render(symx[i]+x1,symy[i]+y1,symbol,0x80,0x80,0x80,0x80,0,0);
    }
    return;
  }
  VIFDynamicDMA.StartDirect();
  VIFDynamicDMA.StartAD();
  /* Transparent background */
  VIFDynamicDMA.AddAD(RGBAQ_SET(0,0,0,0x40,0),RGBAQ);
  VIFDynamicDMA.AddAD(GS_PRIM(GS_SPRITE,0,0,0,1,0,0,0,0), PRIM);
  VIFDynamicDMA.EndAD();
  VIFDynamicDMA.Add128(GS_GIFTAG_BATCH(1,1,0,0,GIF_FLG_REGLIST,
                       GS_BATCH_2(GIF_REG_XYZ2,GIF_REG_XYZ2)));
  VIFDynamicDMA.Add64(XYZ2_SET(((x1+2048)<<4)-8,((y1+2048)<<4)-8,0xFFFFFC));
  VIFDynamicDMA.Add64(XYZ2_SET(((x2+2048)<<4)-8,((y2+2048)<<4)-8,0xFFFFFC));
  if(scurrent>=0) {
    /* Selection background */
    VIFDynamicDMA.StartAD();
    VIFDynamicDMA.AddAD(RGBAQ_SET(0x80,0,0,0x40,0),RGBAQ);
    VIFDynamicDMA.AddAD(GS_PRIM(GS_SPRITE,0,0,0,1,0,0,0,0), PRIM);
    VIFDynamicDMA.EndAD();
    VIFDynamicDMA.Add128(GS_GIFTAG_BATCH(1,1,0,0,GIF_FLG_REGLIST,
                         GS_BATCH_2(GIF_REG_XYZ2,GIF_REG_XYZ2)));
    VIFDynamicDMA.Add64(XYZ2_SET(((x1+2048)<<4)-8,((scurrent*16+y1+2048)<<4)-8,
                                 0xFFFFFD));
    VIFDynamicDMA.Add64(XYZ2_SET(((x2+2048)<<4)-8,
                                 ((scurrent*16+y1+2064)<<4)-8,0xFFFFFD));
  }
  VIFDynamicDMA.EndDirect();
  speed++;
  if((scrolling)&(speed>SCROLL_SPEED)) {
    speed=0;
    scroll++;
    if(scroll>15) {
      scroll=0;
      cursor++;
      if(cursor>=MAX_LINES) cursor=0;
    }
  }
  if(cursor==current) {
    scrolling=false;
    return;
  }
  loc=cursor;
  wa=true;
  cy=y1;
  while(wa) {
    if(loc==cursor) {
      font->Render(x1,cy,lines[loc],0x80,0x80,0x80,0x80,scroll);
      cy=cy+16-scroll;
    }
    else {
      cy+=16;
      es=0;
      if(cy>y2) {
        es=y2-cy;
        wa=false;
      }
      font->Render(x1,cy-16,lines[loc],0x80,0x80,0x80,0x80,es);
    }
    loc++;
    if(loc>=MAX_LINES) loc=0;
    if(loc==current) wa=false;
  } 
}
